/** @file

hippodraw::QtFileDialog implementation

Copyright (C) 2005, 2006  The Board of Trustees of The Leland Stanford
Junior University.  All Rights Reserved.

$Id: QtFileDialog.cxx,v 1.51 2007/07/02 18:17:16 pfkeb Exp $

*/

#ifdef HAVE_CONFIG_H
// for have root cfitsio
#include "config.h"
#endif

#include "QtFileDialog.h"

#include "CanvasWindow.h"
#include "WindowController.h"
#include "Inspector.h"
#include "ListDialog.h"

#include "datasrcs/DataSourceController.h"
#include "datasrcs/DataSource.h"
#include "datasrcs/NTupleController.h"

#ifdef HAVE_CFITSIO
#include "fits/FitsController.h"
#endif

#ifdef HAVE_ROOT
#include "root/RootController.h"
#endif

#include <qmessagebox.h>


using std::exception;
using std::runtime_error;
using std::string;
using std::vector;

using namespace hippodraw;

/* Use `=' syntax so that Doxygen picks it up. */
const string QtFileDialog::s_doc_suffix = ".hpo";
const string QtFileDialog::s_data_suffix = ".tnt";
string QtFileDialog::s_open_filter = "";

QtFileDialog::
QtFileDialog ()
{
}

const string &
QtFileDialog::
createOpenFilter ()
{
  const string doc ( s_doc_suffix );
  const string xml ( ".xml" );
  const string tnt ( ".tnt" );

#if __GNUC__ < 3
#define STRING_CLEAR_DEFECT
#endif

#ifdef STRING_CLEAR_DEFECT
  s_open_filter.erase();
#else
  s_open_filter.clear();
#endif

  s_open_filter = "All files (";
  s_open_filter += "*";
  s_open_filter += ");;";

  s_open_filter += "Documents (*";
  s_open_filter += doc;
  s_open_filter += " *" + xml;

  s_open_filter += ");;Text ntuples (*";
  s_open_filter += tnt;
  s_open_filter += ")";

#ifdef HAVE_ROOT
  const string root ( ".root" );
  s_open_filter += ";;ROOT ntuples(*";
  s_open_filter += root;
  s_open_filter += ")";
#endif

#ifdef HAVE_CFITSIO
  const string fits ( ".fits" );
  s_open_filter += ";;FITS ntuples(*";
  s_open_filter += fits;
  s_open_filter += ")";

  const string gzfits ( ".fits.gz" );
  s_open_filter += ";;Compressed FITS ntuples(*";
  s_open_filter += gzfits;
  s_open_filter += ")";
#endif

  return s_open_filter;
}

const string &
QtFileDialog::
createExportFilter ()
{
#ifdef STRING_CLEAR_DEFECT
  s_open_filter.erase();
#else
  s_open_filter.clear();
#endif

  s_open_filter += "Text ntuples (*";
  s_open_filter += ".tnt";
  s_open_filter += ")";

#ifdef HAVE_CFITSIO
  const string fits ( ".fits" );
  s_open_filter += ";;FITS ntuples(*";
  s_open_filter += fits;
  s_open_filter += ")";

  const string gzfits ( ".fits.gz" );
  s_open_filter += ";;Compressed FITS ntuples(*";
  s_open_filter += gzfits;
  s_open_filter += ")";
#endif

  return s_open_filter;
}


const string &
QtFileDialog::
createBrowseFilter ()
{
  const string doc ( s_doc_suffix );
  const string xml ( ".xml" );
  const string tnt ( ".tnt" );

#if __GNUC__ < 3
#define STRING_CLEAR_DEFECT
#endif

#ifdef STRING_CLEAR_DEFECT
  s_open_filter.erase();
#else
  s_open_filter.clear();
#endif
  s_open_filter = "Documents (*";
  s_open_filter += doc;
  s_open_filter += " *" + xml;
  s_open_filter += ")";

  s_open_filter += ";;All files (";
  s_open_filter += "*.*";
  s_open_filter += ")";

  return s_open_filter;
}

const std::string &
QtFileDialog::
getDocSuffix ()
{
  return s_doc_suffix;
}

const std::string &
QtFileDialog::
getTextSuffix ()
{
  return s_data_suffix;
}

bool
QtFileDialog::
isDocSuffix ( const std::string & suffix )
{
  return suffix == s_doc_suffix || suffix == ".xml";
}

bool
QtFileDialog::
isTextSuffix ( const std::string & suffix )
{
  return suffix == s_data_suffix;
}

bool
QtFileDialog::
isRootSuffix ( const std::string & suffix )
{
  return suffix == ".root";
}

bool
QtFileDialog::
isFitsSuffix ( const std::string & suffix )
{
  return suffix == ".fits";
}

bool
QtFileDialog::
isZippedFitsSuffix ( const std::string & suffix)
{
  return suffix == ".gz";
}

std::string
QtFileDialog::
getExportTupleFilename( QWidget * widget )
{
  const string & filter = createExportFilter ();
  QString filename;

  while (true)
    {
      string tmp;
      QString extname = 0;
      filename 
#if QT_VERSION < 0x040000
	= QFileDialog::getSaveFileName ( QString::null, // starting directory
#else
	= Q3FileDialog::getSaveFileName ( QString::null, // starting directory
#endif
					 filter.c_str(),
					 widget, // parent
					 "export", // name
					 "Save data source to file ...",
					 & extname ); //selected filter

     if ( filename == QString::null ) return tmp;

      if (( filename.endsWith ( ".tnt" ) == false ) &&
	  ( filename.endsWith ( ".fits" ) == false ) &&
	  ( filename.endsWith ( ".fits.gz" )==false ))
					 
	{
	  // Add default suffix if not end with .tnt or .fits.
	  if ( extname == "FITS ntuples(*.fits)" )
	    {
	      filename+=".fits";
	    }
	  else if (extname == "Compressed FITS ntuples(*.fits.gz)")
	    {
	      filename+=".fits.gz";
	    }
	  else
	    { 
	      filename+=".tnt";
	    }
	}     

      QFileInfo info ( filename );
      filename = info.absFilePath();
      
      bool yes = info.exists ();
      if ( yes == false ) break;
      
      string message ( "File exists. \n\n" );
      message += "Over write existing file?";
      int result = QMessageBox::warning ( widget, // parent
					  "Warning", // caption
					  message.c_str (), //
					  QMessageBox::Yes,
					  QMessageBox::No );
      if ( result == QMessageBox::Yes ) break;
    }

  string name = filename.latin1();

  return name;
}
      

const string &
QtFileDialog::
createDocumentFilter ()
{
  s_open_filter.erase();
  s_open_filter = "Doucuments (*";
  s_open_filter += s_doc_suffix.c_str();
  s_open_filter += ")";

  return s_open_filter;
}

void
QtFileDialog::
openTextTuple ( const std::string & filename )
{
  NTupleController * controller = NTupleController::instance ();
  try {
        DataSource * tuple =
    controller -> createNTuple ( filename );
  }
  catch ( const std::exception & e )
    {

      QString message( "Unable to open file\n" );
      message.append(e.what ());

      QMessageBox::critical ( 0, // parent
			      "HippoDraw",
			      message,
			      QMessageBox::Ok,
			      Qt::NoButton );
    }

}

bool
QtFileDialog::
isFitsFile ( const std::string & fn )
{
  bool yes = false;
#ifdef HAVE_CFITSIO
  FitsController * controller = FitsController::instance ();
  FitsFile * fits_file = 0;
  try {
    fits_file = controller ->openFile ( fn );
   if ( fits_file != 0 ) yes = true;
  }
  catch ( ... ) {
    // ignore, was not FITS file
  }
#endif

  return yes;
}

void
QtFileDialog::
  checkDuplicateLabels ( const DataSource * source, QWidget * parent )
{
  const vector < string > & dups = source -> getDuplicatedLabels ();
  if ( dups.empty () == false ) {
    QString message ( "This data source contains two or more columns with\n"
		      "the same label.  The offending labels were\n\n" );
    for ( unsigned int i = 0; i < dups.size(); i++ ) {
      message += dups[i].c_str();
      message += "\n";
    }
    message += "\nThe first column encountered will be used";
    QMessageBox::information (  parent, // parent
			       "HippoDraw", // caption
			       message,
			       QMessageBox::Ok,
			       Qt::NoButton );
  }
}

void
QtFileDialog::
openFitsTuple ( const std::string & filename, QWidget * parent )
{
#ifdef HAVE_CFITSIO
  unsigned int index = 0;

  FitsController * controller = FitsController::instance ();
  const vector < string > & names
    = controller -> getNTupleNames ( filename );

  if ( names.size () > 1 ) {
    if ( names.size () >= 2  ) {
      ListDialog * dialog = new ListDialog ( parent ); // take defaults for rest
      dialog -> setNames ( names );
      int retval = dialog -> exec ();

      if ( retval == QDialog::Accepted ) {
	index = dialog -> selectedItem ();
      }
      delete dialog;
      if ( retval == QDialog::Rejected ) return;
    }
    else { // was your typical empty image HDU
      index = 1;
    }
  }

  try {
    DataSource * ds = 
      controller -> createNTuple ( filename, names [ index ], index );
    checkDuplicateLabels ( ds, parent );
  }
  catch ( const runtime_error & e ) {
    QMessageBox::information ( parent, // parent
			       "HippoDraw", // caption
			       e.what (),
			       QMessageBox::Ok,
			       Qt::NoButton );
    return;
  }

#endif
}

void
QtFileDialog::
openRootTuple ( const std::string & filename, QWidget * parent )
{
#ifdef HAVE_ROOT
  unsigned int index = 0;

  RootController * controller = RootController::instance ();
  const vector < string > & tree_names
    = controller -> getNTupleNames ( filename );
  if ( tree_names.size () > 1 ) {
    ListDialog * dialog = new ListDialog ( parent ); // take defaults for rest
    dialog -> setNames ( tree_names );

    int retval = dialog -> exec ();

    if ( retval == QDialog::Rejected ) return;
    index = dialog -> selectedItem ();
    delete dialog;
  }
  if ( tree_names.empty () ) {
    string message ( "There were no ntuple objects found in this file." );
    message += "\nFile not opened.";

    QMessageBox::information ( parent,
			       "HippoDraw", // caption
			       message.c_str(),
			       QMessageBox::Ok,
			       Qt::NoButton );
    return;
  }

  try {
    DataSource * tuple =
      controller -> createNTuple ( filename, tree_names[index] );
    checkDuplicateLabels ( tuple, parent );
  }
  catch ( const runtime_error & e ) {
    QMessageBox::information ( parent,
			       "HippoDraw", // caption
			       e.what (),
			       QMessageBox::Ok,
			       Qt::NoButton );
    return;
  }

#endif
}

#ifdef HAVE_CFITSIO
void
QtFileDialog::
saveFitsTuple ( const std::string & filename, QWidget * parent )
{
  const string tuple_name = 
    WindowController::instance () 
    -> getInspector() 
    -> getSelectedDataSourceName ();

  try {
    FitsController::instance ()
      ->writeNTupleToFile ( tuple_name, filename );
  }
  catch ( const std::exception & e ) {
    QString message ( "An error occurred in writing file.\n\n" );
    message += e.what();
    QMessageBox::critical ( parent,
			    QString ( "Write error" ), // caption
			    message,
			    QMessageBox::Ok,
			    QMessageBox::NoButton,
			    QMessageBox::NoButton );
  }
}
#endif

void
QtFileDialog::
saveTextTuple ( const std::string & filename, QWidget * parent )
{
  const string tuple_name = 
    WindowController::instance () 
    -> getInspector() 
    -> getSelectedDataSourceName ();

  try {
    NTupleController::instance ()
      ->writeNTupleToFile ( tuple_name, filename );
  }

  catch ( const std::exception & e ) {
    QString message ( "An error occurred in writing file.\n\n" );
    message += e.what();
    QMessageBox::critical ( parent,
			    QString ( "Write error" ), // caption
			    message,
			    QMessageBox::Ok,
			    QMessageBox::NoButton,
			    QMessageBox::NoButton );
  }
}
